package sw

import (
	"context"
	"fmt"
	"github.com/redis/go-redis/v9"
	"gitlab.wanzhuangkj.com/tush/xpkg/gin/xctx"
	"go.uber.org/zap"
	"strconv"
	"sync"
	"time"
)

type RedisDatastore struct {
	client redis.Cmdable
	ttl    time.Duration
	logger *zap.Logger
}

func (d *RedisDatastore) fullKey(key string, start int64) string {
	return fmt.Sprintf("%s@%d", key, start)
}

func (d *RedisDatastore) Add(ctx context.Context, key string, start, value int64) (int64, error) {
	k := d.fullKey(key, start)
	startTime := time.Now()
	c, err := d.client.IncrBy(ctx, k, value).Result()
	if err != nil {
		if d.logger != nil {
			d.logger.Error(fmt.Sprintf("[SlideWindow] add fail,[k:%s add:%d]", k, value), zap.String("err", err.Error()), zap.String("cost", time.Since(startTime).String()), xctx.CtxTraceIDField(ctx))
		}
		return 0, err
	}
	_, _ = d.client.Expire(ctx, k, d.ttl).Result()
	if d.logger != nil {
		d.logger.Info(fmt.Sprintf("[SlideWindow] add success,[k:%s add:%d -> %d]", k, value, c), zap.String("cost", time.Since(startTime).String()), xctx.CtxTraceIDField(ctx))
	}
	return c, err
}

func (d *RedisDatastore) Get(ctx context.Context, key string, start int64) (int64, error) {
	k := d.fullKey(key, start)
	startTime := time.Now()
	value, err := d.client.Get(ctx, k).Result()
	if err != nil {
		if err == redis.Nil {
			err = nil
		}
		if d.logger != nil {
			d.logger.Error(fmt.Sprintf("[SlideWindow] get fail [k:%s]", k), zap.String("err", err.Error()), zap.String("cost", time.Since(startTime).String()), xctx.CtxTraceIDField(ctx))
		}
		return 0, err
	}
	if d.logger != nil {
		d.logger.Info(fmt.Sprintf("[SlideWindow] get success [k:%s v:%s]", k, value), zap.String("cost", time.Since(startTime).String()), xctx.CtxTraceIDField(ctx))
	}
	return strconv.ParseInt(value, 10, 64)
}

func NewRedisDatastore(client redis.Cmdable, ttl time.Duration, logger *zap.Logger) *RedisDatastore {
	return &RedisDatastore{client: client, ttl: ttl, logger: logger}
}

type IPRateLimiter struct {
	mu      sync.Mutex
	options *IPRateLimiterOptions
	store   *RedisDatastore
	ips     map[string]*SLimiter
}

type IPRateLimiterOptions struct {
	windowSize    time.Duration
	maxRequests   int64
	redisAddr     string
	redisDB       int
	redisPassword string
	logger        *zap.Logger
	redisCli      redis.Cmdable
}

func defaultOptions() *IPRateLimiterOptions {
	return &IPRateLimiterOptions{
		windowSize:    time.Minute,
		maxRequests:   60 * 10,
		redisAddr:     "localhost:6379",
		redisPassword: "",
		redisDB:       0,
	}
}

func (x *IPRateLimiterOptions) apply(opts ...Option) {
	for _, opt := range opts {
		opt(x)
	}
}

func WithWindowSize(windowSize time.Duration) Option {
	return func(o *IPRateLimiterOptions) {
		o.windowSize = windowSize
	}
}

func WithMaxRequests(maxRequests int64) Option {
	return func(o *IPRateLimiterOptions) {
		o.maxRequests = maxRequests
	}
}

func WithRedisAddr(redisAddr string) Option {
	return func(o *IPRateLimiterOptions) {
		o.redisAddr = redisAddr
	}
}
func WithRedisDB(redisDB int) Option {
	return func(o *IPRateLimiterOptions) {
		o.redisDB = redisDB
	}
}

func WithRedisPassword(pwd string) Option {
	return func(o *IPRateLimiterOptions) {
		o.redisPassword = pwd
	}
}

func WithLogger(log *zap.Logger) Option {
	return func(o *IPRateLimiterOptions) {
		o.logger = log
	}
}

func WithRedisCli(redisCli redis.Cmdable) Option {
	return func(o *IPRateLimiterOptions) {
		o.redisCli = redisCli
	}
}

type Option func(o *IPRateLimiterOptions)

func NewIPRateLimiter(opts ...Option) *IPRateLimiter {
	options := defaultOptions()
	options.apply(opts...)
	rl := &IPRateLimiter{
		mu:      sync.Mutex{},
		options: options,
		ips:     make(map[string]*SLimiter),
	}
	if options.redisCli != nil {
		rl.store = NewRedisDatastore(options.redisCli, 2*options.windowSize, options.logger)
	} else {
		rl.store = NewRedisDatastore(
			redis.NewClient(&redis.Options{
				Addr:     options.redisAddr,
				Password: options.redisPassword,
				DB:       options.redisDB,
			}),
			2*options.windowSize,
			options.logger,
		)
	}
	return rl
}

type SLimiter struct {
	L    *Limiter
	Stop func()
}

func (x *IPRateLimiter) Allow(ctx context.Context, ip string) bool {
	x.mu.Lock()
	defer x.mu.Unlock()
	if v, ok := x.ips[ip]; ok {
		return v.L.Allow(ctx)
	} else {
		lim, stop := NewLimiter(x.options.windowSize, x.options.maxRequests, func() (Window, StopFunc) {
			return NewSyncWindow(fmt.Sprintf("SyncRateLimiter_%s", ip), NewBlockingSynchronizer(x.store, 500*time.Millisecond))
		})
		x.ips[ip] = &SLimiter{
			L:    lim,
			Stop: stop,
		}
		return lim.Allow(ctx)
	}
}

func (x *IPRateLimiter) Stop() {
	x.mu.Lock()
	defer x.mu.Unlock()
	for _, l := range x.ips {
		l.Stop()
	}
}
